/*
 * Copyright 2017 The Netty Project
 *
 * The Netty Project licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package io.netty.channel;

import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.GenericFutureListener;
import org.junit.Test;
import org.mockito.Mockito;

public class DelegatingChannelPromiseNotifierTest {
    @Test
    public void varargsNotifiersAllowed() {
        ChannelPromise promise = Mockito.mock(ChannelPromise.class);
        DelegatingChannelPromiseNotifier promiseNotifier = new DelegatingChannelPromiseNotifier(promise);

        GenericFutureListener<? extends Future<? super Void>> gfl =
                (GenericFutureListener<? extends Future<? super Void>>) Mockito.mock(GenericFutureListener.class);
        promiseNotifier.addListeners(gfl);
        promiseNotifier.removeListeners(gfl);

        Mockito.verify(promise).addListeners(gfl);
        Mockito.verify(promise).removeListeners(gfl);
    }
}
